package mpp;

import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class LazyPQueue<T> implements PQueue<T> {
	private volatile Node head,tail;
	
	public LazyPQueue() {
		head=new Node(null,Integer.MIN_VALUE);
		tail=new Node(null,Integer.MAX_VALUE);
		head.next=tail;
	}
	
	public void add(T item, int score) {
		while (true){
			Node pred=head;
			Node curr=head.next;
			while (curr.key<score){
				pred=curr;
				curr=curr.next;
			}
			try{
				pred.lock();
				curr.lock();
				if (validate(pred,curr)){					
					Node node=new Node(item,score);
					node.next=curr;
					pred.next=node;
					return;
				}
			}finally{
					curr.unlock();
					pred.unlock();
			}
		}
	}
	
	public boolean contains(T item) {
		Node curr = head;
		while (curr!=tail && !curr.item.equals(item))
			curr = curr.next;
		return curr!=tail && curr.item.equals(item);
	}
	
	public T removeMin() {
		while (true) {
			Node curr = head.next;
			if (head.next==tail)
				return null;
			try {
				head.lock();
				curr.lock();
				if (validate(head, curr)) { //head is always unchecked
					curr.marked = true;
					head.next = curr.next;
					return curr.item;
				}
			} finally {
				curr.unlock();
				head.unlock();
			}
		}
	}
	
	private boolean validate(Node pred, Node curr) {
		return !pred.marked && !curr.marked && pred.next == curr;
	}
	
	private class Node{
		T item;
		int key;
		volatile Node next;
		volatile boolean marked;
		private Lock lock=new ReentrantLock();
		
		void lock(){
			lock.lock();
		}
		
		void unlock(){
			lock.unlock();
		}
		
		Node(T item,int key){
			this.item=item;
			this.key=key;
		}
	}

}
