package mpp;

import java.util.concurrent.atomic.*;

public class LockFreePQueue<T> implements PQueue<T> {

	volatile Node head,tail;
	
	LockFreePQueue(){
		head=new Node(null,Integer.MIN_VALUE);
		tail=new Node(null,Integer.MAX_VALUE);
		tail.next=new AtomicMarkableReference<Node>(null,false);
		head.next=new AtomicMarkableReference<Node>(tail,false);
	}
	
	public void add(T item, int score) {
		 while (true) {
			 Window window = find(head, score);
			 Node pred = window.pred, curr = window.curr;
			 if (curr.key == score) {
				 break;
			 } else {
				 Node node = new Node(item,score);
				 node.next = new AtomicMarkableReference<Node>(curr, false);
				 if (pred.next.compareAndSet(curr, node, false, false)) 
					 break; 
			 }
		 }
	}

	public T removeMin() {
		Boolean snip; 
		while (true) {
			Window window=find(head,head.key);
			Node pred = window.pred; 
			Node curr = window.curr;
			T item=curr.item;
			if (curr==tail) {
				return null;
			} else {
				Node succ = curr.next.getReference();
				snip = curr.next.compareAndSet(succ, succ, false, true);
				if (!snip) 
					continue;
				pred.next.compareAndSet(curr, succ, false, false);
				return item;
			}
		}
	}
	
	public Window find(Node head, int key) {
		 Node pred = null, curr = null, succ = null;
		 boolean[] marked = {false}; 
		 boolean snip;
		 retry: while (true) {
			 	pred = head;
			 	curr = pred.next.getReference();
			 	while (true) {
					succ = curr.next.get(marked); 
					while (marked[0]) {
						snip = pred.next.compareAndSet(curr,succ, false, false);
						   if (!snip) 
							   continue retry;
						   curr = succ; 
						   succ = curr.next.get(marked); 
					}
					if (curr.key >= key)
						return new Window(pred, curr);
					pred = curr;
					curr = succ; 
			 	}
			}
	}
	
	public boolean contains(T item,int key) {
		boolean[] marked = {false}; 
	    Node curr = this.head;
	    while (curr.key < key)
	    	curr = curr.next.getReference();
	    /*Node succ =*/ curr.next.get(marked);
	    return (curr.key == key && !marked[0]);
	  } 


	private class Window{
		public Node pred, curr;
		Window(Node myPred, Node myCurr) {
			pred = myPred; curr = myCurr;
		}
	}
	
	private class Node{
		T item;
		int key;
		AtomicMarkableReference<Node> next;

		Node(T item,int key){
			this.item=item;
			this.key=key;
		}
	}
}
